!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
!                                                              
!     Apparent Horizon Finder for Kerr BH                      
!                                    in isotropic coordinates  
!                                                              
!--------------------------------------------------------------
!                                                              
!     Sample Code for Lecture in NR/HEP2 Spring School         
!                                                              
!                                     Coded by Hirotada Okawa  
!                                                              
!==============================================================
!   How to compile and use this program in terminal(bash)
!==============================================================
!   $ gfortran -O2 -ffast-math -o ahf_kbh sor_AHF_KBH_ISO.f90
!   $ ./ahf_kbh
!--------------------------------------------------------------
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

module inc_coord
  implicit none

!--------------------------------------------------------------
!     Grid points
!--------------------------------------------------------------
  integer,parameter :: jli=1
  integer,parameter :: jui=80
  integer,parameter :: jlb=jli-1
  integer,parameter :: jub=jui+1

!--------------------------------------------------------------
!     Maximum/Minimum Coordinates (physical)
!--------------------------------------------------------------
  real(8),parameter :: pi=atan(1.d0)*4.d0
  real(8),parameter :: xlower=0.
  real(8),parameter :: xupper=pi
  real(8),parameter :: dx=(xupper-xlower)/dble(jub-jli)
  real(8),parameter :: dxi=1.d0/dx

!--------------------------------------------------------------
!     Array for Coordinates (physical)
!--------------------------------------------------------------
  real(8),dimension(jlb:jub) :: x

!--------------------------------------------------------------
!     Variables to solve
!--------------------------------------------------------------
  real(8),dimension(jlb:jub) :: h, hprev

!--------------------------------------------------------------
!     Source term for Poisson equation
!--------------------------------------------------------------
  real(8),dimension(jlb:jub) :: src
end module inc_coord

program main
  use inc_coord
  implicit none

!--------------------------------------------------------------
!     Definition of parameters
!--------------------------------------------------------------
  integer,parameter :: stepmax=1d8
  real(8),parameter :: errormax=1.d-10
  real(8),parameter :: fpar=1.d0
  real(8),parameter :: omega=1.7d0
  real(8),parameter :: mm=1.d0, aa=0.8d0

!--------------------------------------------------------------
!     Definition of temporary variables to use
!--------------------------------------------------------------
  integer :: j, step
  real(8) :: xx, hh, hx, hxx, hhi, h1
  real(8) :: errortmp, vtmp
  real(8) :: cosx, sinx

  real(8) :: rbl, del, Ad, sig, ci2 
  real(8) :: grr, gtt, gitt, girr
  real(8) :: rblr, sigr, sigt, delr, Adr, Adt
  real(8) :: girrr, girrt, gittr, gittt

!--------------------------------------------------------------
!     Output File
!--------------------------------------------------------------
  open(200,file='data_kbh_iso.dat')

!--------------------------------------------------------------
!     Initialization
!--------------------------------------------------------------
  do j=jlb,jub
     x(j)  = xlower +(dble(j)-0.5d0)*dx    ! Coordinates
     ! h(j)  = 1.d0                        ! initial guess
     h(j)  = dcos(xx)                      ! initial guess
     hprev(j) = h(j)                       ! previous variable
  end do

!--------------------------------------------------------------
!     Main Loop
!--------------------------------------------------------------
  do step=0,stepmax

!--------------------------------------------------------------
!     Preserve data of previous step
!--------------------------------------------------------------
     do j=jlb,jub
        hprev(j) = h(j)
     end do

!--------------------------------------------------------------
!     SOR Method
!--------------------------------------------------------------
     do j=jli,jui

!==============================================================
!     Definition of Source term
!==============================================================
        xx  = x(j)
        sinx= dsin(xx)
        cosx= dcos(xx)
        hh  = h(j)
        hhi = 1.d0/hh
        h1  = 1.d0 +hhi**2
        hx  = (h(j+1) -h(j-1))*0.5d0*dxi
        hxx = (h(j+1) +h(j-1) -2.d0*hh)*dxi**2

        rbl = hh*(1.d0 +(mm +aa)*0.5d0*hhi)*(1.d0 +(mm -aa)*0.5d0*hhi)
        del = rbl**2 -2.d0*mm*rbl +aa**2
        Ad  = (rbl**2 +aa**2)**2 -del*aa**2*sinx**2
        sig = rbl**2 +aa**2*cosx**2

        grr = sig*hhi**2
        gtt = sig
        gitt= 1.d0/sig
        girr= hh**2*gitt
        ci2 = girr +gitt*hx**2

        rblr = 1.d0 -(mm**2-aa**2)*0.25d0*hhi**2
        sigr = 2.d0*rbl*rblr
        sigt =-2.d0*aa**2*cosx*sinx
        delr = 2.d0*(rbl-mm)*rblr
        Adr  = 4.d0*(rbl**2 +aa**2)*rbl*rblr -delr*aa**2*sinx**2
        Adt  =-2.d0*del*aa**2*sinx*cosx
        girrr= hh*(2.d0*sig -hh*sigr)/sig**2
        girrt=-hh**2*sigt/sig**2
        gittr=-sigr/sig**2
        gittt=-sigt/sig**2

        src(j) = hxx -hh*ci2*girr +0.5d0*hh**2*ci2/sig*(sigr*girr-sigt*gitt*hx) &
&              +0.5d0*hh**2*ci2/Ad*(Adr*girr-Adt*gitt*hx) &
&              -hh**2*ci2*gitt*cosx/sinx*hx +hh**2*ci2*girrr &
&              -hh**2*ci2*hx*gittt -hh**2*ci2*gitt*hxx &
&              -0.5d0*hh**2*girr*(girrr +gittr*hx**2) &
&              +0.5d0*hh**2*gitt*hx*(girrt +gittt*hx**2+2.d0*gitt*hx*hxx)


        h(j) = (1.-omega)*h(j)+omega*0.5d0*( hprev(j+1) +h(j-1) -dx**2*src(j) )
     end do

!==============================================================
!     Impose Boundary Condition
!==============================================================
     h(jub)=h(jui)         ! Neumann Boundary Condition
     h(jlb)=h(jli)         ! Neumann Boundary Condition

!--------------------------------------------------------------
!     Check if values converge
!--------------------------------------------------------------
     errortmp=0.d0
     vtmp=0.d0
     do j=jli,jui
        errortmp = errortmp +(h(j)-hprev(j))*(h(j)-hprev(j))*dx**2
        vtmp = vtmp + dx**2
     end do
     errortmp = dsqrt(errortmp/vtmp)
     if( (errortmp.le.errormax) .and. (step.gt.1) ) exit

!--------------------------------------------------------------
!     Next Guess(Mix Previous value) 
!--------------------------------------------------------------
     do j=jlb,jub
        h(j) = fpar*h(j) +(1.d0-fpar)*hprev(j)
     end do

     write(*,*) "Step=",step,"Error=",errortmp
  end do

!--------------------------------------------------------------
!     Print Data
!--------------------------------------------------------------
  do j=jlb,jub
     write(200,'(4e16.8e2)') x(j),h(j),hprev(j),src(j)
  end do

!--------------------------------------------------------------
!     End of Program
!--------------------------------------------------------------
  write(*,*) "End of Run",errortmp
  close(200)

end program main
